# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import subprocess
import sys

__version__ = '0.2.4'

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--profile', required=True, help='The AWS profile to configure')

    parser.add_argument('--sso-start-url')
    parser.add_argument('--sso-region')

    interactive_group = parser.add_mutually_exclusive_group()
    interactive_group.add_argument('--set-auth-interactive', '-i', action='store_const', const=True, dest='interactive', help='Enable interactive auth')
    interactive_group.add_argument('--set-auth-noninteractive', '-n', action='store_const', const=False, dest='interactive', help='Disable interactive auth')

    parser.add_argument('--version', action='store_true')

    args = parser.parse_args()

    if args.version:
        print(__version__)
        parser.exit()

    def get(name):
        return subprocess.run(['aws', 'configure', 'get', f'profile.{args.profile}.{name}'], capture_output=True).stdout

    def set(name, value):
        subprocess.run(['aws', 'configure', 'set', f'profile.{args.profile}.{name}', value or ''], capture_output=True, check=True)

    if args.sso_start_url:
        set('sso_start_url', args.sso_start_url)
    elif not get('sso_start_url') and os.environ.get('AWS_CONFIGURE_SSO_DEFAULT_SSO_START_URL'):
        set('sso_start_url', os.environ.get('AWS_CONFIGURE_SSO_DEFAULT_SSO_START_URL'))

    if args.sso_region:
        set('sso_region', args.sso_region)
    elif not get('sso_region') and os.environ.get('AWS_CONFIGURE_SSO_DEFAULT_SSO_REGION'):
        set('sso_region', os.environ.get('AWS_CONFIGURE_SSO_DEFAULT_SSO_REGION'))

    result = subprocess.run(f'aws configure sso --profile {args.profile}', shell=True).returncode

    if result:
        sys.exit(result)

    if os.environ.get('AWS_CONFIGURE_SSO_DISABLE_CREDENTIAL_PROCESS', '').lower() not in ['1', 'true']:
        credential_process_opts = ''
        if args.interactive is True:
            credential_process_opts += ' --interactive'
        elif args.interactive is False:
            credential_process_opts += ' --noninteractive'
        set('credential_process', f'aws-sso-credential-process --profile {args.profile}{credential_process_opts}')

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aws_sso_credential_process']

package_data = \
{'': ['*']}

install_requires = \
['botocore>=1.16.0',
 'docutils>=0.10,<0.16',
 'jmespath>=0.7.1,<1.0.0',
 'python-dateutil>=2.1,<3.0.0',
 'urllib3>=1.20,<1.26']

entry_points = \
{'console_scripts': ['aws-configure-sso-profile = '
                     'aws_sso_credential_process.aws_configure_sso_profile:main',
                     'aws-sso-configure-profile = '
                     'aws_sso_credential_process.aws_configure_sso_profile:main',
                     'aws-sso-credential-process = '
                     'aws_sso_credential_process.aws_sso_credential_process:main']}

setup_kwargs = {
    'name': 'aws-sso-credential-process',
    'version': '0.2.5',
    'description': 'Bring AWS SSO-based credentials to the AWS SDKs until they have proper support',
    'long_description': '# aws-sso-credential-process\n**Get AWS SSO working with all the SDKs that don\'t understand it yet**\n\nCurrently, [AWS SSO](https://aws.amazon.com/single-sign-on/) support is implemented in the [AWS CLI v2](https://aws.amazon.com/blogs/developer/aws-cli-v2-is-now-generally-available/), but the capability to usage the credentials retrieved from AWS SSO by the CLI v2 has not been implemented in the various AWS SDKs. However, they all support the [credential process](https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sourcing-external.html) system. This tool bridges the gap by implementing a credential process provider that understands the SSO credential retrieval and caching system. Once AWS implements the necessary support in the SDK for your favorite language, this tool will no longer be necessary.\n\nIf you try this and your tools still don\'t work with the credentials, you can get the credentials themselves using [aws-export-credentials](https://github.com/benkehoe/aws-export-credentials), which can also inject them as environment variables for your program.\n\n## Quickstart\n\n1. I recommend you install [`pipx`](https://pipxproject.github.io/pipx/), which installs the tool in an isolated virtualenv while linking the script you need.\n\nMac:\n```bash\nbrew install pipx\npipx ensurepath\n```\n\nOther:\n```bash\npython3 -m pip install --user pipx\npython3 -m pipx ensurepath\n```\n\n2. Install the tool.\n```bash\npipx install aws-sso-credential-process\n```\n\n3. Set up your `.aws/config` file for AWS SSO as normal (you can use `aws configure sso --profile my-sso-profile` to do this as well):\n\n```\n[profile my-sso-profile]\n\nregion = us-east-2\noutput = yaml\n\nsso_start_url = https://something.awsapps.com/start\nsso_region = us-east-2\nsso_account_id = 123456789012\nsso_role_name = MyLeastPrivilegeRole\n```\n\n4. Then, just add a [`credential_process` entry](https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sourcing-external.html) to the profile, using the `--profile` flag with the same profile name:\n\n```\n[profile my-sso-profile]\n\ncredential_process = aws-sso-credential-process --profile my-sso-profile\n\nregion = us-east-2\noutput = yaml\n\nsso_start_url = https://something.awsapps.com/start\nsso_region = us-east-2\nsso_account_id = 123456789012\nsso_role_name = MyLeastPrivilegeRole\n\n```\n\n5. You\'re done! Test it out:\n```bash\naws sso login --profile my-sso-profile\npython -c "import boto3; print(boto3.Session(profile_name=\'my-sso-profile\').client(\'sts\').get_caller_identity())"\n```\n\n## Configuration\n\nThe order of configuration matches the AWS CLI and SDKs: values from CLI parameters take precedence, followed by env vars, followed by settings in `.aws/config`.\n\nThe `--profile` parameter doesn\'t work like the same parameter on the AWS CLI, and cannot be set from the environment; it\'s intended only to help make the `credential_process` entry in a profile more concise.\n\nThe `aws-configure-sso-profile` tool wraps `aws configure sso` to help you set up profiles in `.aws/config`; you can set the environment variables `AWS_CONFIGURE_SSO_DEFAULT_SSO_START_URL` and `AWS_CONFIGURE_SSO_DEFAULT_SSO_REGION` to set defaults for those values so you\'re not typing them all the time. The tool will set up the `credential_process` entry as well. Note that `--profile` is required (unlike `aws configure sso`).\n\n### Interactive authentication\n\nThe most important thing to determine is whether or not you want to allow interactive authentication, which is off by default (so that the behavior is the same as the AWS CLI v2).\n\nWhen interactive authentication is off, you need to use the CLI v2\'s `aws sso login` to login through AWS SSO. If you haven\'t logged in or your session has expired, the process will fail and interrupt whatever you\'re doing.\n\nWith interactive authentication turned on, the functionality of `aws sso login` will be triggered automatically; a browser will pop up to prompt you to log in (or, if you\'re already logged in, it will prompt you to approve the login). This is useful when you\'re running scripts interactively, but bad for automated processes that are incapable of logging in.\n\n**To enable interactive authentication, the best way is to set `AWS_SSO_INTERACTIVE_AUTH=true` in your environment.** This lets you control whether interactive auth is enabled for a given profile depending on the situation you\'re using it for. Otherwise, you can set `sso_interactive_auth=true` in your profile in `.aws/config`, or use the `--interactive` flag for the process. Note that you can use the `--noninteractive` flag to disable interactive auth even if the environment variable is set.\n\nNote that if you\'ve got your profile set up as shown above, the AWS CLI v2 won\'t get interactive authentication, because it will natively use the profile configuration, skipping this tool as a credential process. If you really want interactive auth with the CLI, you can put the AWS SSO configuration information as parameters to the tool in the credential process directive, instead of directly in the profile, and then the CLI will use credential process as well.\n\n### Debugging\nSetting the `--debug` flag or the env var `AWS_SSO_CREDENTIAL_PROCESS_DEBUG=true` will cause debug output to be sent to `.aws/sso/aws-sso-credential-process-log.txt`. Note that this file will contain your credentials, though these credentials are both short-lived and also cached within the same directory.\n\n### Account\n\n* `.aws/config`: `sso_account_id`\n* env var: `AWS_SSO_ACCOUNT_ID`\n* parameter: `--account-id`\n\n### Role\n\n* `.aws/config`: `sso_role_name`\n* env var: `AWS_SSO_ROLE_NAME`\n* parameter: `--role-name`\n\n### SSO Start URL\n\n* `.aws/config`: `sso_start_url`\n* env var: `AWS_SSO_START_URL`\n* parameter: `--start-url`\n\n### SSO Region\n\n* `.aws/config`: `sso_region`\n* env var: `AWS_SSO_REGION`\n* parameter: `--region`\n',
    'author': 'Ben Kehoe',
    'author_email': 'ben@kehoe.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/benkehoe/aws-sso-credential-process',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This code is based on the code for the AWS CLI v2's `aws sso login` functionality
# https://github.com/aws/aws-cli/tree/v2/awscli/customizations/sso

import argparse
import os
import sys
import webbrowser
import json
import traceback
import logging
import datetime
import io
import textwrap
import collections

#TODO: use getpass functionality to get around captured stdout & stderr
def print_tty(message):
    print(message, file=sys.stderr)
def prompt_tty(message):
    raise NotImplementedError

from botocore.session import Session
from botocore.credentials import JSONFileCache
from botocore.exceptions import ClientError

from .utils import SSOTokenFetcher
from .credentials import SSOCredentialFetcher

__version__ = '0.2.5'

class InvalidSSOConfigError(Exception):
    pass

class AuthDispatchError(Exception):
    pass

class InteractiveAuthDisabledError(Exception):
    pass

# https://gist.github.com/benkehoe/c61337ddb0c213bb35d05aaa8fad2577
BotoErrorInfo = collections.namedtuple('BotoErrorInfo', ['code', 'message', 'http_status_code', 'operation_name'])
def get_error_info(client_error):
    return BotoErrorInfo(client_error.response.get('Error', {}).get('Code'), client_error.response.get('Error', {}).get('Message'), client_error.response.get('ResponseMetadata', {}).get('HTTPStatusCode'), client_error.operation_name)

def boto_error_matches(client_error, *args):
    errs = tuple(e for e in (client_error.response.get('Error', {}).get('Code'), client_error.response.get('Error', {}).get('Message'), client_error.response.get('ResponseMetadata', {}).get('HTTPStatusCode'), client_error.operation_name) if e is not None)
    return any(arg in errs for arg in args)

SSO_TOKEN_DIR = os.path.expanduser(
    os.path.join('~', '.aws', 'sso', 'cache')
)

CREDENTIALS_CACHE_DIR = os.path.expanduser(
    os.path.join('~', '.aws', 'sso', 'cache')
)

LOG_FILE = os.path.expanduser(
    os.path.join('~', '.aws', 'sso', 'aws-sso-credential-process-log.txt')
)

LOGGER = logging.getLogger(__name__)

CONFIG_VARS = [
    ('start url', 'sso_start_url'),
    ('SSO region', 'sso_region'),
    ('role', 'sso_role_name'),
    ('account', 'sso_account_id'),
    ('interactive', 'sso_interactive_auth')
]

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--profile', help='Extract settings from the given profile')

    parser.add_argument('--role-name')
    parser.add_argument('--account-id')
    parser.add_argument('--start-url')
    parser.add_argument('--region')

    interactive_group = parser.add_mutually_exclusive_group()
    interactive_group.add_argument('--interactive', '-i', action='store_const', const=True, dest='interactive', help='Enable interactive auth')
    interactive_group.add_argument('--noninteractive', '-n', action='store_const', const=False, dest='interactive', help='Disable interactive auth')

    parser.add_argument('--force-refresh', action='store_true', help='Do not reuse cached AWS SSO token')
    parser.add_argument('--debug', action='store_true', help='Write to the debugging log file')

    parser.add_argument('--version', action='store_true')

    args = parser.parse_args()

    if args.version:
        print(__version__)
        parser.exit()

    if args.debug or os.environ.get('AWS_SSO_CREDENTIAL_PROCESS_DEBUG', '').lower() in ['1', 'true']:
        logging.basicConfig(level=logging.DEBUG, filename=LOG_FILE, filemode='w')
    else:
        logging.disable(logging.CRITICAL)

    LOGGER.info(f'Starting credential process at {datetime.datetime.now().isoformat()}')

    if args.role_name is None and os.environ.get('AWS_SSO_ROLE_NAME'):
        LOGGER.debug(f"Using role from env: {os.environ['AWS_SSO_ROLE_NAME']}")
        args.role_name = os.environ['AWS_SSO_ROLE_NAME']

    if args.account_id is None and os.environ.get('AWS_SSO_ACCOUNT_ID'):
        LOGGER.debug(f"Using acccount from env: {os.environ['AWS_SSO_ACCOUNT_ID']}")
        args.account_id = os.environ['AWS_SSO_ACCOUNT_ID']

    # if args.role_name and args.role_name.startswith('arn'):
    #     parts = args.role_name.split(':')
    #     args.account_id = parts[4]
    #     args.role_name = parts[5].split('/', 1)[1]

    if args.start_url is None and os.environ.get('AWS_SSO_START_URL'):
        args.start_url = os.environ['AWS_SSO_START_URL']

    if args.region is None and os.environ.get('AWS_SSO_REGION'):
        args.region = os.environ['AWS_SSO_REGION']

    if args.interactive is None:
        if os.environ.get('AWS_SSO_INTERACTIVE_AUTH'):
            LOGGER.debug(f"Setting interactive auth from env: {os.environ['AWS_SSO_INTERACTIVE_AUTH']}")
            args.interactive = os.environ['AWS_SSO_INTERACTIVE_AUTH'].lower() in ['true', '1']
        else:
            args.interactive = False

    session_kwargs = {}

    if args.profile:
        session_kwargs['profile'] = args.profile

    arg_config = {
        'sso_start_url': args.start_url,
        'sso_region': args.region,
        'sso_role_name': args.role_name,
        'sso_account_id': args.account_id,
        'sso_interactive_auth': 'true' if args.interactive else 'false',
    }

    LOGGER.info(f'CONFIG FROM ARGS: {json.dumps(arg_config)}')

    try:
        session = Session(**session_kwargs)

        if args.profile:
            profile_config = session.get_scoped_config()
            LOGGER.info(f'CONFIG FROM PROFILE: {json.dumps(profile_config)}')
        else:
            profile_config = {}

        config = get_config(arg_config, profile_config)

        LOGGER.info(f'CONFIG: {json.dumps(config)}')

        interactive = config['sso_interactive_auth'].lower() == 'true'

        token_loader = get_token_loader(
            session=session,
            sso_region=config['sso_region'],
            interactive=interactive,
            force_refresh=args.force_refresh,
        )

        if not config['sso_account_id']:
            #TODO: if interactive, prompt for account
            raise InvalidSSOConfigError('Missing account id')

        if not config['sso_role_name']:
            #TODO: if interactive, prompt for role
            raise InvalidSSOConfigError('Missing role')

        credentials = get_credentials(
            session=session,
            sso_region=config['sso_region'],
            start_url=config['sso_start_url'],
            account_id=config['sso_account_id'],
            role_name=config['sso_role_name'],
            token_loader=token_loader
        )

        output = {
            "Version": 1,
            "AccessKeyId": credentials['access_key'],
            "SecretAccessKey": credentials['secret_key'],
            "SessionToken": credentials['token'],
            "Expiration": credentials['expiry_time'].replace('UTC', 'Z')
        }
        LOGGER.debug('CREDENTIALS: ' + json.dumps(output))

        print(json.dumps(output, separators=(',', ':')))
    except InteractiveAuthDisabledError as e:
        LOGGER.info('Auth needed but interactive auth disabled')
        print('Interactive auth disabled, use `aws sso login` and try again', file=sys.stderr)
        sys.exit(1)
    except InvalidSSOConfigError as e:
        LOGGER.error(e)
        print(e, file=sys.stderr)
        sys.exit(2)
    except AuthDispatchError as e:
        LOGGER.error(e)
        print(e, file=sys.stderr)
        sys.exit(3)
    except ClientError as e:
        LOGGER.error(e, exc_info=True)
        #TODO: print a different message for AccessDeniedException during CreateToken? -> user canceled login
        # boto_error_matches(e, 'CreateToken', 'AccessDeniedException')
        print('ERROR:', e, file=sys.stderr)
        sys.exit(4)
    except Exception as e:
        LOGGER.error(e, exc_info=True)
        print('ERROR:', e, file=sys.stderr)
        sys.exit(5)

def get_config(arg_config, profile_config):
    sso_config = {}
    missing_vars = []
    for friendly_name, config_var_name in CONFIG_VARS:
        if arg_config.get(config_var_name):
            sso_config[config_var_name] = arg_config[config_var_name]
        elif config_var_name not in profile_config:
            missing_vars.append((friendly_name, config_var_name))
            sso_config[config_var_name] = None
        else:
            sso_config[config_var_name] = profile_config[config_var_name]

    required_vars = ['sso_start_url', 'sso_region', 'sso_account_id', 'sso_role_name']

    # TODO: enable when interactive account and role picker is implemented
    # interactive = sso_config.get('sso_interactive_auth') == 'true'
    # if interactive:
    #     required_vars = ['sso_start_url', 'sso_region']

    missing_requred_vars = [v[0] for v in missing_vars if v[1] in required_vars]
    if missing_requred_vars:
        raise InvalidSSOConfigError(
            'Missing ' + ', '.join(missing_requred_vars)
        )
    return sso_config

class OpenBrowserHandler(object):
    def __init__(self, outfile=None, open_browser=None):
        self._outfile = outfile or sys.stderr
        if open_browser is None:
            open_browser = webbrowser.open_new_tab
        self._open_browser = open_browser

    def __call__(self, userCode, verificationUri,
                 verificationUriComplete, **kwargs):
        message = textwrap.dedent(f"""\
        AWS SSO login required.
        Attempting to open the SSO authorization page in your default browser.
        If the browser does not open or you wish to use a different device to
        authorize this request, open the following URL:

        {verificationUri}

        Then enter the code:

        {userCode}
        """)

        print_tty(message)

        if self._open_browser:
            try:
                return self._open_browser(verificationUriComplete)
            except InteractiveAuthDisabledError:
                raise
            except Exception as e:
                raise AuthDispatchError('Failed to open browser') from e
                # LOG.debug('Failed to open browser:', exc_info=True)

def _non_interactive_auth_raiser(*args, **kwargs):
    raise InteractiveAuthDisabledError

def get_token_loader(session, sso_region, interactive=False, token_cache=None,
                 on_pending_authorization=None, force_refresh=False):

    if token_cache is None:
        token_cache = JSONFileCache(SSO_TOKEN_DIR)

    if on_pending_authorization is None:
        if interactive:
            on_pending_authorization = OpenBrowserHandler(
                outfile=sys.stderr,
                open_browser=webbrowser.open_new_tab,
            )
        else:
            on_pending_authorization = _non_interactive_auth_raiser

    token_fetcher = SSOTokenFetcher(
        sso_region=sso_region,
        client_creator=session.create_client,
        cache=token_cache,
        on_pending_authorization=on_pending_authorization,
    )

    def token_loader(start_url):
        token_response = token_fetcher.fetch_token(
            start_url=start_url,
            force_refresh=force_refresh
        )
        LOGGER.debug(f'TOKEN RESPONSE: {token_response}')
        return token_response['accessToken']

    return token_loader


def get_credentials(session, sso_region, start_url, account_id, role_name, token_loader, cache=None):

    if cache is None:
        cache = JSONFileCache(CREDENTIALS_CACHE_DIR)

    credential_fetcher = SSOCredentialFetcher(
        start_url=start_url,
        sso_region=sso_region,
        role_name=role_name,
        account_id=account_id,
        client_creator=session.create_client,
        cache=cache,
        token_loader=token_loader,
    )

    return credential_fetcher.fetch_credentials()
